import collections

from .statements import (
    Select,
    From,
    Where,
    GroupBy,
    Having,
    OrderBy,
)
from .utils import clean_query


class OracleQueryBuilder:
    COMPONENT_ORDER = {
        1: Select,
        2: From,
        3: Where,
        4: GroupBy,
        5: Having,
        6: OrderBy
    }

    def __init__(self):
        self.statements = {}

    def __str__(self):
        return self.build_query()

    @property
    def query(self):
        return self.build_query()

    @property
    def plain_query(self):
        return clean_query(self.build_query())

    def select(self, *args):
        """
        This method will receive arguments for columns
        OracleQueryBuilder object
        :param args: Columns to select
            example:
                with the following arguments
                    args = FOO, BAR
                the parsed query will be
                    SELECT
                        FOO,
                        BAR

        :return: An OracleQueryBuilder object
        """
        if not self.statements.get(Select):
            self.statements[Select] = Select()

        self.statements[Select].add(*args)

        return self

    def from_table(self, *args):
        """
            This method will call From object to add the
             args into the list of tables
        :param args: Tables for the FROM statement of the query
        :return: An OracleQueryBuilder object
        """
        if not self.statements.get(From):
            self.statements[From] = From()

        self.statements[From].add(*args)

        return self

    def where(self, *args):
        """
            This method will call Where object to add
            args into the list of constraints
        :param args: Constraints for the Where statement of the query
        :return: An OracleQueryBuilder object
        """
        if not self.statements.get(Where):
            self.statements[Where] = Where()

        self.statements[Where].add(*args)

        return self

    def build_query(self):
        query = ''
        ordered_statements = collections.OrderedDict(sorted(self.COMPONENT_ORDER.items()))
        for _, statement in ordered_statements.items():
            statement_instance = self.statements.get(statement)
            if statement_instance:
                query += statement_instance.parse()

        return query



import re


class QueryUtils:
    def clean_query(self, query):
        query = self.remove_sql_commentaries(query)
        return self.remove_duplicate_spaces(self.remove_line_and_tabs(query))

    @staticmethod
    def remove_sql_commentaries(query):
        if not query:
            return ''

        # remove all occurance singleline comments (//COMMENT\n ) from query
        return re.sub(re.compile("--.*?\n"), "", query)

    @staticmethod
    def remove_line_and_tabs(string):
        return string.replace('\n', ' ').replace('\t', ' ').strip()

    @staticmethod
    def remove_duplicate_spaces(string):
        return re.sub(' +', ' ', string)


def clean_query(query):
    return QueryUtils().clean_query(query)



from pyqybe.exceptions import InvalidOperator

EQUAL_OPERATOR_STR = 'equal'
BIGGER_OPERATOR_STR = 'bigger'
BIGGER_OR_EQUAL_OPERATOR_STR = 'bigger or equal'
SMALLER_OPERATOR_STR = 'smaller'
SMALLER_OR_EQUAL_OPERATOR_STR = 'smaller or equal'
DIFFERENT_OPERATOR_STR = 'different'
IN_OPERATOR_STR = 'in'
LIKE_OPERATOR_STR = 'like'
BETWEEN_OPERATOR_STR = 'between'
ALL_OPERATORS_STR = [EQUAL_OPERATOR_STR, BIGGER_OPERATOR_STR, BIGGER_OR_EQUAL_OPERATOR_STR, SMALLER_OPERATOR_STR,
                 SMALLER_OR_EQUAL_OPERATOR_STR, DIFFERENT_OPERATOR_STR, IN_OPERATOR_STR, LIKE_OPERATOR_STR,
                 BETWEEN_OPERATOR_STR,
                 ]


class Operator:
    @staticmethod
    def like(value):
        """
            LIKE operator
                NAME LIKE '%WORKER%'
        :param value: the 2nd term of the expression
            For the example above, it would be '%WORKER%'
        :return: tuple containing the operator string and the 2nd term of the expression
        """
        return LIKE_OPERATOR_STR, value

    @staticmethod
    def equal(value):
        """
            Equality operator
                CITY == 'Uberaba'
        :param value: the 2nd term of the expression
            For the example above, it would be 'Uberaba'
        :return: tuple containing the operator string and the 2nd term of the expression
        """
        return EQUAL_OPERATOR_STR, value

    @staticmethod
    def bigger(value):
        return BIGGER_OPERATOR_STR, value

    @staticmethod
    def bigger_or_equal(value):
        return BIGGER_OR_EQUAL_OPERATOR_STR, value

    @staticmethod
    def smaller(value):
        return SMALLER_OPERATOR_STR, value

    @staticmethod
    def smaller_or_equal(value):
        return SMALLER_OR_EQUAL_OPERATOR_STR, value

    @staticmethod
    def different(value):
        return DIFFERENT_OPERATOR_STR, value

    @staticmethod
    def between(first, second):
        """
            BETWEEN operator
                REFERENCE_DATE BETWEEN '20170701' AND '20170731'
        :param first: the 1st value of the 2nd term of the expression
            For the example above, it would be '20170701'
        :param second: the 1st value of the 2nd term of the expression
            For the example above, it would be '20170731'
        :return: tuple containing the operator string and the 2nd term of the expression
        """
        return BETWEEN_OPERATOR_STR, (first, second)

    @staticmethod
    def in_operator(*values):
        """
            IN operator
                COUNTRY IN ('BRAZIL', 'NETHERLANDS', 'FINLAND')
        :param values: the 2nd term of the expression
            Lists and tuples are accepted
            For the example above, it would be either:
                ['BRAZIL', 'NETHERLANDS', 'FINLAND']
                or 'BRAZIL', 'NETHERLANDS', 'FINLAND'
        :return: tuple containing the operator string and the 2nd term of the expression
        """
        operator_values = []
        for value in values:
            if isinstance(value, list):
                operator_values.extend(value)
            else:
                operator_values.append(value)

        return IN_OPERATOR_STR, operator_values

    @staticmethod
    def sniff(*expression):
        """
            sniff will the operator for the expression
        :param expression: The expression to be looked up
            Examples:
                i) 1,
                ii) 1, 2, 3
                iii) [1, 2, 3]
                iv) equal(5)
        :return: tuple containing the operator string and the expression
        """

        if len(expression) == 1:
            expression = expression[0]

            if isinstance(expression, tuple):
                if expression[0] in ALL_OPERATORS_STR:
                    return expression[0], expression[1]
                else:
                    raise InvalidOperator(expression)

            if isinstance(expression, list):
                return IN_OPERATOR_STR, expression

            return EQUAL_OPERATOR_STR, expression

        else:
            if isinstance(expression, tuple):
                return IN_OPERATOR_STR, [el for el in expression]


class OperatorParser:
    # Templates
    EQUAL_TEMPLATE = '{} == {}'
    BIGGER_TEMPLATE = '{} > {}'
    BIGGER_OR_EQUAL_TEMPLATE = '{} >= {}'
    SMALLER_TEMPLATE = '{} < {}'
    SMALLER_OR_EQUAL_TEMPLATE = '{} <= {}'
    DIFFERENT_TEMPLATE = '{} <> {}'
    IN_TEMPLATE = '{} IN ({})'
    LIKE_TEMPLATE = '{} LIKE {}'
    BETWEEN_TEMPLATE = '{} BETWEEN {} AND {}'

    def __init__(self, operator=None):
        self.operator = operator

        self.parsers = {
            EQUAL_OPERATOR_STR: self.equal,
            BIGGER_OPERATOR_STR: self.bigger,
            BIGGER_OR_EQUAL_OPERATOR_STR: self.bigger_or_equal,
            SMALLER_OPERATOR_STR: self.smaller,
            SMALLER_OR_EQUAL_OPERATOR_STR: self.smaller_or_equal,
            DIFFERENT_OPERATOR_STR: self.different,
            IN_OPERATOR_STR: self.in_operator,
            LIKE_OPERATOR_STR: self.like,
            BETWEEN_OPERATOR_STR: self.between
        }

    def parse(self, column, *args):
        return self.parsers[self.operator](column, *args)

    def equal(self, column, value):
        return self.EQUAL_TEMPLATE.format(
            column,
            self.parse_type(value)
        )

    def bigger(self, column, value):
        return self.BIGGER_TEMPLATE.format(
            column,
            self.parse_type(value)
        )

    def bigger_or_equal(self, column, value):
        return self.BIGGER_OR_EQUAL_TEMPLATE.format(
            column,
            self.parse_type(value)
        )

    def smaller(self, column, value):
        return self.SMALLER_TEMPLATE.format(
            column,
            self.parse_type(value)
        )

    def smaller_or_equal(self, column, value):
        return self.SMALLER_OR_EQUAL_TEMPLATE.format(
            column,
            self.parse_type(value)
        )

    def different(self, column, value):
        return self.DIFFERENT_TEMPLATE.format(
            column,
            self.parse_type(value)
        )

    def in_operator(self, column, values):
        return self.IN_TEMPLATE.format(
            column,
            ', '.join(self.parse_type(value) for value in values),
        )

    def like(self, column, value):
        return self.LIKE_TEMPLATE.format(
            column,
            self.parse_type(value)
        )

    def between(self, column, *values):
        # TODO: Find a better way for this approach: values = values[0]
        values = values[0]
        return self.BETWEEN_TEMPLATE.format(
            column,
            self.parse_type(values[0]),
            self.parse_type(values[1])
        )

    @staticmethod
    def parse_type(value):
        if isinstance(value, str):
            return '\'{}\''.format(value)

        return str(value)



from .expressions import (
    Ex,
    ExOr
)
from .operators import Operator
from .oracle_query_builder import OracleQueryBuilder
from .statements import (
    Select,
    Where,
    OrderBy,
    From,
    GroupBy,
    Having,
    Statement
)

ex = ex_and = Ex
ex_or = ExOr
op = Operator
query = OracleQueryBuilder

__author__ = 'Eryk Humberto Oliveira Alves'
__email__ = 'erykwho@gmail.com'
__version__ = '0.0.3'



from pyqybe.operators import Operator, OperatorParser


class Expression(list):
    def __init__(self, *expressions, order=None, join_str):
        super().__init__()
        self._expressions = []
        self.join_str = join_str
        self.parse_expressions(*expressions, order=order)

    def parse_expressions(self, *expressions, order):
        equations = []
        for expression in expressions:
            if isinstance(expression, list):
                equations.extend(expression)
            else:
                equations.extend(self._parse_expression(expression, order))

        self.extend(equations)

    def add(self, *expressions):
        self.extend(*expressions)
        return self

    @staticmethod
    def _parse_expression(expression, order=None):
        equations = []
        if not order:
            order = expression.keys()

        for column in order:
            operator, value = Operator.sniff(expression[column])
            equations.append(OperatorParser(operator).parse(column, value))

        return equations

    def extend(self, equations):
        self_contain = True if len(equations) > 1 else False

        formatted_equation = ' {} '.format(self.join_str).join(equations)
        if self_contain:
            formatted_equation = '({})'.format(formatted_equation)

        super().append(formatted_equation)


class Ex(Expression):
    JOIN_STR = 'AND'

    def __init__(self, *expressions, order=None):
        super().__init__(*expressions, order=order, join_str=self.JOIN_STR)


class ExOr(Expression):
    JOIN_STR = 'OR'

    def __init__(self, *expressions, order=None):
        super().__init__(*expressions, order=order, join_str=self.JOIN_STR)



class Statement:
    def __init__(self, statement_str='', *elements, sep=', '):
        self.sep = sep
        self.statement_str = statement_str
        self._elements = list(elements)

    def add(self, *elements):
        self._elements.extend(elements)
        return self

    def parse(self):
        if not self._elements:
            return ''

        statement_str = '{statement} {elements} '.format(
            statement=self.statement_str,
            elements='{sep} '.format(sep=self.sep).join(
                '{}'.format(value) for value in self._elements)
        )
        return statement_str


class Select(Statement):
    STATEMENT_STR = 'SELECT'

    def __init__(self, *elements):
        super().__init__(self.STATEMENT_STR, *elements)


class Where(Statement):
    STATEMENT_STR = 'WHERE'
    SEP = ' AND '

    def __init__(self, *elements):
        super().__init__(self.STATEMENT_STR, *elements, sep=self.SEP)

    def add(self, elements):
        if not isinstance(elements, list):
            elements = [elements]

        self._elements.extend(elements)
        return self

class From(Statement):
    STATEMENT_STR = 'FROM'

    def __init__(self, *elements):
        super().__init__(self.STATEMENT_STR, *elements)


class GroupBy(Statement):
    pass


class Having(Statement):
    pass


class OrderBy(Statement):
    pass



import re

from setuptools import setup


def version():
    version_file_name = "pyqybe/__init__.py"
    with open(version_file_name, 'rt') as version_file:
        version_file_content = version_file.read()

        version_regex = r"^__version__ = ['\"]([^'\"]*)['\"]"
        match = re.search(version_regex, version_file_content, re.M)
        if match:
            return match.group(1)
        else:
            raise RuntimeError("Unable to find version string in %s." % (version_file_name,))


setup(name='pyqybe',
      version=version(),
      author='Eryk Humberto Oliveira Alves',
      author_email='erykwho@gmail.com',
      url='https://github.com/otrabalhador/pyqybe/',
      packages=['pyqybe'],
      keywords='python query sql builder pyqybe',
      classifiers=[
          'Programming Language :: Python :: 3',
          'Programming Language :: PL/SQL',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Topic :: Scientific/Engineering :: Information Analysis',
          'Topic :: Scientific/Engineering :: Mathematics',
          'Operating System :: POSIX',
      ],
      test_suite="pyqybe.tests",)
